package org.warp.commonutils.locks;

import it.cavallium.concurrentlocks.ReadWriteUpdateLock;
import java.util.Objects;
import java.util.Set;
import org.jetbrains.annotations.NotNull;

/**
 * Fine-grained locking on data or on the entire collection for a snapshottable collection
 * @param <T>
 */
public class SnapshottableCollectionLock<T> {

	private final TransitReadWriteLock fullLock;
	private final TransitReadWriteLock dataLock;
	private final Striped<ReadWriteUpdateLock> transactionLock;

	public SnapshottableCollectionLock(int stripedSize) {
		this.transactionLock = Striped.readWriteUpdateLock(stripedSize);
		this.fullLock = new TransitReadWriteLock();
		this.dataLock = new TransitReadWriteLock();
	}

	public enum DataLockType {
		READ,
		/**
		 * "Update" allow reads but blocks writes and other updates.
		 */
		UPDATE,
		WRITE
	}

	public enum FullLockType {
		/** <p>This lock is a read lock for "full", nothing for "data"</p>
		 * <p>Useful for coarse iterators</p>
		 *
		 * <ul>
		 * <li>full<ul>
		 *     <li>read: allow</li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * <li>data<ul>
		 *     <li>read/update: allow</li>
		 *     <li>write: allow</li>
		 * </ul></li>
		 * </ul>
		 */
		DENY_ONLY_FULL_WRITE,
		/** <p>This lock is a read lock for "full" and a read lock for "data"</p>
		 *
		 * <ul>
		 * <li>full<ul>
		 *     <li>read: allow</li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * <li>data<ul>
		 *     <li>read/update: allow</li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * </ul>
		 */
		ALLOW_ONLY_READ,
		/** <p>This lock is a write lock for "full" and a read lock for "data"</p>
		 *
		 * <ul>
		 * <li>full<ul>
		 *     <li>read: <u>deny</u></li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * <li>data<ul>
		 *     <li>read/update: allow</li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * </ul>
		 */
		ALLOW_ONLY_ITEMS_READ,
		/** <p>This lock is a read lock for "full" and a write lock for "data"</p>
		 *
		 * <ul>
		 * <li>full<ul>
		 *     <li>read: allow</li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * <li>data<ul>
		 *     <li>read/update: <u>deny</u></li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * </ul>
		 */
		ALLOW_ONLY_FULL_READ,
		/** <p>This lock is a write lock for "full" and a write lock for "data"</p>
		 *
		 * <ul>
		 * <li>full<ul>
		 *     <li>read: <u>deny</u></li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * <li>data<ul>
		 *     <li>read/update: <u>deny</u></li>
		 *     <li>write: <u>deny</u></li>
		 * </ul></li>
		 * </ul>
		 */
		EXCLUSIVE;

		public static final Set<FullLockType> READ_FULL_TYPES = Set.of(
				DENY_ONLY_FULL_WRITE,
				ALLOW_ONLY_READ,
				ALLOW_ONLY_FULL_READ
		);
		public static final Set<FullLockType> WRITE_FULL_TYPES = Set.of(ALLOW_ONLY_ITEMS_READ, EXCLUSIVE);
		public static final Set<FullLockType> READ_DATA_TYPES = Set.of(ALLOW_ONLY_READ, ALLOW_ONLY_ITEMS_READ);
		public static final Set<FullLockType> WRITE_DATA_TYPES = Set.of(ALLOW_ONLY_FULL_READ, EXCLUSIVE);
		public static final Set<FullLockType> NOOP_DATA_TYPES = Set.of(DENY_ONLY_FULL_WRITE);
	}

	public void fullLock(Object snapshot, FullLockType fullLockMode) {
		if (snapshot == null) {
			// Transit on full locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_FULL_READ:
					fullLock.startTransitRead();
					break;
				case ALLOW_ONLY_ITEMS_READ:
				case EXCLUSIVE:
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
			// Manage full locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_FULL_READ:
					fullLock.disallowTransitWrite();
					break;
				case ALLOW_ONLY_ITEMS_READ:
				case EXCLUSIVE:
					fullLock.disallowTransit();
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
			// Manage data locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
					break;
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_ITEMS_READ:
					dataLock.disallowTransitWrite();
					break;
				case ALLOW_ONLY_FULL_READ:
				case EXCLUSIVE:
					dataLock.disallowTransit();
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
		}
	}

	public void fullUnlock(Object snapshot, FullLockType fullLockMode) {
		if (snapshot == null) {
			// Manage full locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_FULL_READ:
					fullLock.reAllowTransitWrite();
					break;
				case ALLOW_ONLY_ITEMS_READ:
				case EXCLUSIVE:
					fullLock.reAllowTransit();
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
			// Transit on full locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_FULL_READ:
					fullLock.endTransitRead();
					break;
				case ALLOW_ONLY_ITEMS_READ:
				case EXCLUSIVE:
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
			// Manage data locks
			switch (fullLockMode) {
				case DENY_ONLY_FULL_WRITE:
					break;
				case ALLOW_ONLY_READ:
				case ALLOW_ONLY_ITEMS_READ:
					dataLock.reAllowTransitWrite();
					break;
				case ALLOW_ONLY_FULL_READ:
				case EXCLUSIVE:
					dataLock.reAllowTransit();
					break;
				default:
					throw new IllegalStateException("Unexpected value: " + fullLockMode);
			}
		}
	}

	private void partialLockKey(DataLockType dataLockMode, ReadWriteUpdateLock lock) {
		switch (dataLockMode) {
			case READ:
				dataLock.startTransitRead();
				lock.readLock().lock();
				break;
//todo: check carefully if an update lock can lock dataLock only for data read.
// If it can, replace startTransitWrite with startTransitRead here.
			case UPDATE:
				dataLock.startTransitWrite();
				lock.updateLock().lock();
				break;
			case WRITE:
				dataLock.startTransitWrite();
				lock.writeLock().lock();
				break;
		}
	}

	private void partialUnlockKey(DataLockType dataLockMode, ReadWriteUpdateLock lock) {
		switch (dataLockMode) {
			case READ:
				lock.readLock().unlock();
				dataLock.endTransitRead();
				break;
//todo: check carefully if an update lock can lock dataLock only for data read.
// If it can, replace startTransitWrite with startTransitRead here.
			case UPDATE:
				lock.updateLock().unlock();
				dataLock.endTransitWrite();
				break;
			case WRITE:
				lock.writeLock().unlock();
				dataLock.endTransitWrite();
				break;
		}
	}

	private void partialLockKeyUpgrade(DataLockType dataLockMode, ReadWriteUpdateLock lock) {
		switch (dataLockMode) {
			case READ:
				lock.readLock().lock();
				break;
			case UPDATE:
				lock.updateLock().lock();
				break;
			case WRITE:
				lock.writeLock().lock();
				break;
		}
	}

	private void partialUnlockKeyUpgrade(DataLockType dataLockMode, ReadWriteUpdateLock lock) {
		switch (dataLockMode) {
			case READ:
				lock.readLock().unlock();
				break;
			case UPDATE:
				lock.writeLock().unlock();
				break;
			case WRITE:
				lock.writeLock().unlock();
				break;
		}
	}

	public void dataLock(Object snapshot, DataLockType dataLockMode, @NotNull T key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			var lock = transactionLock.get(key);
			partialLockKey(dataLockMode, lock);
		}
	}

	public void dataLockAt(Object snapshot, DataLockType dataLockMode, int index) {
		if (snapshot == null) {
			var lock = transactionLock.getAt(index);
			partialLockKey(dataLockMode, lock);
		}
	}

	public void dataLockUpgrade(Object snapshot, DataLockType dataLockMode, @NotNull T key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			var lock = transactionLock.get(key);
			partialLockKeyUpgrade(dataLockMode, lock);
		}
	}

	public void dataLockUpgradeAt(Object snapshot, DataLockType dataLockMode, int index) {
		if (snapshot == null) {
			var lock = transactionLock.getAt(index);
			partialLockKeyUpgrade(dataLockMode, lock);
		}
	}

	public void dataUnlock(Object snapshot, DataLockType dataLockMode, @NotNull T key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			var lock = transactionLock.get(key);
			partialUnlockKey(dataLockMode, lock);
		}
	}

	public void dataUnlockAt(Object snapshot, DataLockType dataLockMode, int index) {
		if (snapshot == null) {
			var lock = transactionLock.getAt(index);
			partialUnlockKey(dataLockMode, lock);
		}
	}

	public void dataUnlockUpgrade(Object snapshot, DataLockType dataLockMode, @NotNull T key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			var lock = transactionLock.get(key);
			partialUnlockKeyUpgrade(dataLockMode, lock);
		}
	}

	public void dataUnlockUpgradeAt(Object snapshot, DataLockType dataLockMode, @NotNull int index) {
		if (snapshot == null) {
			var lock = transactionLock.getAt(index);
			partialUnlockKeyUpgrade(dataLockMode, lock);
		}
	}

	public void dataLockMulti(Object snapshot, DataLockType dataLockMode, @NotNull Iterable<? extends T> key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			for (ReadWriteUpdateLock lock : transactionLock.bulkGet(key)) {
				partialLockKey(dataLockMode, lock);
			}
		}
	}

	public void dataLockAtMulti(Object snapshot, DataLockType dataLockMode, @NotNull Iterable<Integer> key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			for (ReadWriteUpdateLock lock : transactionLock.bulkGetAt(key)) {
				partialLockKey(dataLockMode, lock);
			}
		}
	}

	public void dataUnlockMulti(Object snapshot, DataLockType dataLockMode, @NotNull Iterable<? extends T> key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			for (ReadWriteUpdateLock lock : transactionLock.bulkGet(key)) {
				partialUnlockKey(dataLockMode, lock);
			}
		}
	}

	public void dataUnlockAtMulti(Object snapshot, DataLockType dataLockMode, @NotNull Iterable<Integer> key) {
		if (snapshot == null) {
			Objects.requireNonNull(key, () -> "Lock key must not be null");
			for (ReadWriteUpdateLock lock : transactionLock.bulkGetAt(key)) {
				partialUnlockKey(dataLockMode, lock);
			}
		}
	}
}
