package org.warp.commonutils.locks;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.warp.commonutils.locks.SnapshottableCollectionLock.DataLockType;
import org.warp.commonutils.locks.SnapshottableCollectionLock.FullLockType;

public class SnapshottableCollectionLockTest {

	@Test
	public void testNoop() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		checkSituation(lock, true, true, true, true);
	}

	@Test
	public void testDenyOnlyFullWrite() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		lock.fullLock(null, FullLockType.DENY_ONLY_FULL_WRITE);
		checkSituation(lock, true, false, true, true);
		lock.fullUnlock(null, FullLockType.DENY_ONLY_FULL_WRITE);
	}

	@Test
	public void testAllowOnlyRead() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		lock.fullLock(null, FullLockType.ALLOW_ONLY_READ);
		checkSituation(lock, true, false, true, false);
		lock.fullUnlock(null, FullLockType.ALLOW_ONLY_READ);
	}

	@Test
	public void testAllowOnlyItemsRead() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		lock.fullLock(null, FullLockType.ALLOW_ONLY_ITEMS_READ);
		checkSituation(lock, false, false, true, false);
		lock.fullUnlock(null, FullLockType.ALLOW_ONLY_ITEMS_READ);
	}

	@Test
	public void testAllowOnlyFullRead() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		lock.fullLock(null, FullLockType.ALLOW_ONLY_FULL_READ);
		checkSituation(lock, true, false, false, false);
		lock.fullUnlock(null, FullLockType.ALLOW_ONLY_FULL_READ);
	}

	@Test
	public void testExclusive() {
		var lock = new SnapshottableCollectionLock<Long>(50);
		lock.fullLock(null, FullLockType.EXCLUSIVE);
		checkSituation(lock, false, false, false, false);
		lock.fullUnlock(null, FullLockType.EXCLUSIVE);
	}

	private void checkSituation(SnapshottableCollectionLock<Long> lock, boolean fullReadAllow, boolean fullWriteAllow, boolean dataReadAllow, boolean dataWriteAllow) {

		var pool = Executors.newWorkStealingPool();

		for (FullLockType readFullType : FullLockType.READ_FULL_TYPES) {
			if (fullReadAllow) {
				Assertions.assertDoesNotThrow(() -> {
					CompletableFuture.runAsync(() -> {
						lock.fullLock(null, readFullType);
						lock.fullUnlock(null, readFullType);
					}, pool).get(2, TimeUnit.SECONDS);
				});
			} else {
				Assertions.assertThrows(TimeoutException.class, () -> {
					CompletableFuture.runAsync(() -> {
						lock.fullLock(null, readFullType);
						lock.fullUnlock(null, readFullType);
					}, pool).get(50, TimeUnit.MILLISECONDS);
				});
			}
		}

		for (FullLockType writeFullType : FullLockType.WRITE_FULL_TYPES) {
			if (fullWriteAllow) {
				Assertions.assertDoesNotThrow(() -> {
					CompletableFuture.runAsync(() -> {
						lock.fullLock(null, writeFullType);
						lock.fullUnlock(null, writeFullType);
					}, pool).get(2, TimeUnit.SECONDS);
				});
			} else {
				Assertions.assertThrows(TimeoutException.class, () -> {
					CompletableFuture.runAsync(() -> {
						lock.fullLock(null, writeFullType);
						lock.fullUnlock(null, writeFullType);
					}, pool).get(50, TimeUnit.MILLISECONDS);
				});
			}
		}

		if (dataReadAllow) {
			Assertions.assertDoesNotThrow(() -> {
				CompletableFuture.runAsync(() -> {
					lock.dataLock(null, DataLockType.READ, 1L);
					lock.dataUnlock(null, DataLockType.READ, 1L);
				}, pool).get(2, TimeUnit.SECONDS);
			});
		} else {
			Assertions.assertThrows(TimeoutException.class, () -> {
				CompletableFuture.runAsync(() -> {
					lock.dataLock(null, DataLockType.READ, 1L);
					lock.dataUnlock(null, DataLockType.READ, 1L);
				}, pool).get(50, TimeUnit.MILLISECONDS);
			});
		}

		if (dataWriteAllow) {
			Assertions.assertDoesNotThrow(() -> {
				CompletableFuture.runAsync(() -> {
					lock.dataLock(null, DataLockType.WRITE, 1L);
					lock.dataUnlock(null, DataLockType.WRITE, 1L);
				}, pool).get(2, TimeUnit.SECONDS);
			});
		} else {
			Assertions.assertThrows(TimeoutException.class, () -> {
				CompletableFuture.runAsync(() -> {
					lock.dataLock(null, DataLockType.WRITE, 1L);
					lock.dataUnlock(null, DataLockType.WRITE, 1L);
				}, pool).get(50, TimeUnit.MILLISECONDS);
			});
		}
	}
}
